import datetime
from typing import Union, List, Dict, Optional, Text, Tuple  # noqa: F401

import requests
from six import text_type, string_types

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import Category, Multiplicity, Classification
from pykechain.exceptions import APIError, IllegalArgumentError, NotFoundError, MultipleFoundError
from pykechain.extra_utils import relocate_model, move_part_instance, relocate_instance, get_mapping_dictionary, \
    get_edited_one_many
from pykechain.models import Scope2
from pykechain.models.property2 import Property2
from pykechain.models.tree_traversal import TreeObject
from pykechain.utils import is_uuid, find


class Part2(TreeObject):
    """A virtual object representing a KE-chain part.

    :ivar id: UUID of the part
    :type id: basestring
    :ivar name: Name of the part
    :type name: basestring
    :ivar ref: Reference of the part (slug of the original name)
    :type ref: basestring
    :ivar description: description of the part
    :type description: basestring or None
    :ivar created_at: the datetime when the object was created if available (otherwise None)
    :type created_at: datetime or None
    :ivar updated_at: the datetime when the object was last updated if available (otherwise None)
    :type updated_at: datetime or None
    :ivar category: The category of the part, either 'MODEL' or 'INSTANCE' (of :class:`pykechain.enums.Category`)
    :type category: basestring
    :ivar parent_id: The UUID of the parent of this part
    :type parent_id: basestring or None
    :ivar properties: The list of :class:`Property2` objects belonging to this part.
    :type properties: List[Property2]
    :ivar multiplicity: The multiplicity of the part being one of the following options: ZERO_ONE, ONE, ZERO_MANY,
                        ONE_MANY, (reserved) M_N (of :class:`pykechain.enums.Multiplicity`)
    :type multiplicity: basestring
    :ivar scope_id: scope UUID of the Part
    :type scope_id: basestring
    :ivar properties: the list of properties of this part
    :type properties: List[AnyProperty]

    Examples
    --------
    For the category property

    >>> bike = project.part('Bike')
    >>> bike.category
    'INSTANCE'

    >>> bike_model = project.model('Bike')
    >>> bike_model.category
    'MODEL'

    >>> bike_model == Category.MODEL
    True
    >>> bike == Category.INSTANCE
    True

    For the multiplicity property

    >>> bike = project.models('Bike')
    >>> bike.multiplicity
    ONE_MANY

    >>> from pykechain.enums import Multiplicity
    >>> bike.multiplicity == Multiplicity.ONE_MANY
    True

    """

    def __init__(self, json: Dict, **kwargs):
        """Construct a part from a KE-chain 2 json response.

        :param json: the json response to construct the :class:`Part` from
        :type json: dict
        """
        # we need to run the init of 'Base' instead of 'Part' as we do not need the instantiation of properties
        super(Part2, self).__init__(json, **kwargs)

        self.scope_id = json.get('scope_id')

        self.ref = json.get('ref')  # type: Text
        self.category = json.get('category')  # type: Category
        self.description = json.get('description')  # type: Text
        self.multiplicity = json.get('multiplicity')  # type: Text
        self.classification = json.get('classification')  # type: Classification

        self.properties = [Property2.create(p, client=self._client)
                           for p in sorted(json['properties'], key=lambda p: p.get('order', 0))]  # type: list

        proxy_data = json.get('proxy_source_id_name', dict())  # type: Optional[Dict]
        self._proxy_model_id = proxy_data.get('id') if proxy_data else None  # type: Optional[Text]

    def refresh(self, json: Optional[Dict] = None, url: Optional[Text] = None, extra_params: Optional[Dict] = None):
        """Refresh the object in place."""
        if extra_params is None:
            extra_params = {}
        extra_params.update(API_EXTRA_PARAMS['part2'])
        super(Part2, self).refresh(json=json,
                                   url=self._client._build_url('part2', part_id=self.id),
                                   extra_params=extra_params)

    #
    # Family and structure methods
    #

    def property(self, name: Text = None) -> 'AnyProperty':
        """Retrieve the property belonging to this part based on its name or uuid.

        :param name: property name, ref or property UUID to search for
        :return: a single :class:`Property`
        :raises NotFoundError: if the `Property` is not part of the `Part`

        Example
        -------
        >>> part = project.part('Bike')
        >>> part.properties
        [<pyke Property ...>, ...]
        # this returns a list of all properties of this part

        >>> gears = part.property('Gears')
        >>> gears.value
        6

        >>> gears = part.property('123e4567-e89b-12d3-a456-426655440000')
        >>> gears.value
        6

        """
        if is_uuid(name):
            found = find(self.properties, lambda p: name == p.id)
        else:
            found = find(self.properties, lambda p: name == p.name or name == p.ref)

        if not found:
            raise NotFoundError("Could not find property with name, ref or id '{}'".format(name))

        return found

    def scope(self) -> 'Scope2':
        """Scope this Part belongs to.

        This property will return a `Scope` object. It will make an additional call to the KE-chain API.

        :return: the scope
        :rtype: :class:`pykechain.models.Scope`
        :raises NotFoundError: if the scope could not be found
        """
        return self._client.scope(pk=self.scope_id, status=None)

    def parent(self) -> 'Part2':
        """Retrieve the parent of this `Part`.

        :return: the parent :class:`Part` of this part
        :raises APIError: if an Error occurs

        Example
        -------
        >>> part = project.part('Frame')
        >>> bike = part.parent()

        """
        return self._client.part(pk=self.parent_id, category=self.category) if self.parent_id else None

    def children(self, **kwargs) -> Union['PartSet', List['Part2']]:
        """Retrieve the children of this `Part` as `PartSet`.

        When you call the :func:`Part.children()` method without any additional filtering options for the children,
        the children are cached to help speed up subsequent calls to retrieve the children. The cached children are
        returned as a list and not as a `Partset`.

        When you *do provide* additional keyword arguments (kwargs) that act as a specific children filter, the
        cached children are _not_ used and a separate API call is made to retrieve only those children.

        :param kwargs: Additional search arguments to search for, check :class:`pykechain.Client.parts`
                       for additional info
        :return: a set of `Parts` as a :class:`PartSet`. Will be empty if no children. Will be a `List` if the
                 children are retrieved from the cached children.
        :raises APIError: When an error occurs.

        Example
        -------
        A normal call, which caches all children of the bike. If you call `bike.children` twice only 1 API call is made.

        >>> bike = project.part('Bike')
        >>> direct_descendants_of_bike = bike.children()

        An example with providing additional part search parameters 'name__icontains'. Children are retrieved from the
        API, not the bike's internal (already cached in previous example) cache.

        >>> bike = project.part('Bike')
        >>> wheel_children_of_bike = bike.children(name__icontains='wheel')

        """
        if not kwargs:
            # no kwargs provided is the default, we aim to cache it.
            if self._cached_children is None:
                self._cached_children = list(self._client.parts(parent=self.id, category=self.category))
            return self._cached_children
        else:
            # if kwargs are provided, we assume no use of cache as specific filtering on the children is performed.
            return self._client.parts(parent=self.id, category=self.category, **kwargs)

    def child(self,
              name: Optional[Text] = None,
              pk: Optional[Text] = None,
              **kwargs) -> 'Part2':
        """
        Retrieve a child object.

        :param name: optional, name of the child
        :type name: str
        :param pk: optional, UUID of the child
        :type: pk: str
        :return: Child object
        :raises MultipleFoundError: whenever multiple children fit match inputs.
        :raises NotFoundError: whenever no child matching the inputs could be found.
        """
        if not (name or pk):
            raise IllegalArgumentError('You need to provide either "name" or "pk".')

        if self._cached_children:
            # First try to find the child without calling KE-chain.
            if name:
                part_list = [c for c in self.children() if c.name == name]
            else:
                part_list = [c for c in self.children() if c.id == pk]
        else:
            part_list = []

        if not part_list:
            # Refresh children list from KE-chain by using a keyword-argument.
            if name:
                part_list = self.children(name=name)
            else:
                part_list = self.children(pk=pk)

        criteria = '\nname: {}\npk: {}\nkwargs: {}'.format(name, pk, kwargs)

        # If there is only one, then that is the part you are looking for
        if len(part_list) == 1:
            part = part_list[0]

        # Otherwise raise the appropriate error
        elif len(part_list) > 1:
            raise MultipleFoundError('{} has more than one matching child.{}'.format(self, criteria))
        else:
            raise NotFoundError('{} has no matching child.{}'.format(self, criteria))
        return part

    def populate_descendants(self, batch: int = 200) -> None:
        """
        Retrieve the descendants of a specific part in a list of dicts and populate the :func:`Part.children()` method.

        Each `Part` has a :func:`Part.children()` method to retrieve the children on the go. This function
        prepopulates the children and the children's children with its children in one call, making the traversal
        through the parttree blazingly fast.

        .. versionadded:: 2.1

        .. versionchanged:: 3.3.2 now populates child parts instead of this part

        :param batch: Number of Parts to be retrieved in a batch
        :type batch: int (defaults to 200)
        :returns: None
        :raises APIError: if you cannot create the children tree.

        Example
        -------
        >>> bike = project.part('Bike')
        >>> bike.populate_descendants(batch=150)

        """
        all_descendants = list(self._client.parts(
            category=self.category,
            batch=batch,
            descendants=self.id,
        ))[1:]  # remove the part itself, which is returned on index 0

        # Create mapping table from a parent part ID to its children
        children_by_parent_id = dict()
        for descendant in all_descendants:
            if descendant.parent_id in children_by_parent_id:
                children_by_parent_id[descendant.parent_id].append(descendant)
            else:
                children_by_parent_id[descendant.parent_id] = [descendant]

        # Populate every descendant with its children
        for descendant in all_descendants:
            if descendant.id in children_by_parent_id:
                descendant._cached_children = children_by_parent_id[descendant.id]
            else:
                descendant._cached_children = []

        self._cached_children = children_by_parent_id.get(self.id, list())

        return None

    def all_children(self) -> List['Part2']:
        """
        Retrieve a flat list of all descendants, sorted depth-first. Also populates all descendants.

        :returns list of child objects
        :rtype List
        """
        self.populate_descendants()
        return super().all_children()

    def siblings(self, **kwargs) -> Union['PartSet', List['Part2']]:
        """Retrieve the siblings of this `Part` as `PartSet`.

        Siblings are other Parts sharing the same parent of this `Part`, including the part itself.

        :param kwargs: Additional search arguments to search for, check :class:`pykechain.Client.parts`
                       for additional info
        :return: a set of `Parts` as a :class:`PartSet`. Will be empty if no siblings.
        :raises APIError: When an error occurs.
        """
        if self.parent_id:
            return self._client.parts(parent=self.parent_id, category=self.category, **kwargs)
        else:
            from pykechain.models.partset import PartSet
            return PartSet(parts=[])

    #
    # Model and Instance(s) methods
    #

    def model(self) -> 'Part2':
        """
        Retrieve the model of this `Part` as `Part`.

        For instance, you can get the part model of a part instance. But trying to get the model of a part that
        has no model, like a part model, will raise a :exc:`NotFoundError`.

        .. versionadded:: 1.8

        :return: the model of this part instance as :class:`Part` with category `MODEL`
        :raises NotFoundError: if no model found

        Example
        -------
        >>> front_fork = project.part('Front Fork')
        >>> front_fork_model = front_fork.model()

        """
        if self.category == Category.INSTANCE:
            model_id = self._json_data.get('model_id')
            return self._client.model(pk=model_id)
        else:
            raise NotFoundError("Part {} has no model".format(self.name))

    def instances(self, **kwargs) -> Union['PartSet', List['Part2']]:
        """
        Retrieve the instances of this `Part` as a `PartSet`.

        For instance, if you have a model part, you can get the list of instances that are created based on this
        moodel. If there are no instances (only possible if the multiplicity is :attr:`enums.Multiplicity.ZERO_MANY`)
        than a :exc:`NotFoundError` is returned

        .. versionadded:: 1.8

        :return: the instances of this part model :class:`PartSet` with category `INSTANCE`
        :raises NotFoundError: if no instances found

        Example
        -------
        >>> wheel_model = project.model('Wheel')
        >>> wheel_instance_set = wheel_model.instances()

        An example with retrieving the front wheels only using the 'name__contains' search argument.

        >>> wheel_model = project.model('Wheel')
        >>> front_wheel_instances = wheel_model.instances(name__contains='Front')

        """
        if self.category == Category.MODEL:
            return self._client.parts(model=self, category=Category.INSTANCE, **kwargs)
        else:
            raise NotFoundError("Part {} is not a model".format(self.name))

    def instance(self) -> 'Part2':
        """
        Retrieve the single (expected) instance of this 'Part' (of `Category.MODEL`) as a 'Part'.

        See :func:`Part.instances()` method for documentation.

        :return: :class:`Part` with category `INSTANCE`
        :raises NotFoundError: if the instance does not exist
        :raises MultipleFoundError: if there are more than a single instance returned
        """
        instances_list = list(self.instances())
        if len(instances_list) == 1:
            return instances_list[0]
        elif len(instances_list) > 1:
            raise MultipleFoundError("Part {} has more than a single instance. "
                                     "Use the `Part.instances()` method".format(self.name))
        else:
            raise NotFoundError("Part {} has no instance".format(self.name))

    #
    # CRUD operations
    #

    def edit(self, name: Optional[Text] = None, description: Optional[Text] = None, **kwargs) -> None:
        """Edit the details of a part (model or instance).

        For an instance you can edit the Part instance name and the part instance description. To alter the values
        of properties use :func:`Part.update()`.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param name: optional name of the part to edit
        :param description: (optional) description of the part
        :type description: basestring or None
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :return: None
        :raises IllegalArgumentError: when the type or value of an argument provided is incorrect
        :raises APIError: in case an Error occurs

        Example
        -------
        For changing a part:

        >>> front_fork = project.part('Front Fork')
        >>> front_fork.edit(name='Front Fork - updated')
        >>> front_fork.edit(name='Front Fork cruizer', description='With my ragtop down so my hair can blow' )

        for changing a model:

        >>> front_fork = project.model('Front Fork')
        >>> front_fork.edit(name='Front Fork basemodel', description='Some description here')

        """
        update_dict = {'id': self.id}
        if name is not None:
            if not isinstance(name, (string_types, text_type)):
                raise IllegalArgumentError("name should be provided as a string")
            update_dict.update({'name': name})
        if description is not None:
            if not isinstance(description, (string_types, text_type)):
                raise IllegalArgumentError("description should be provided as a string")
            update_dict.update({'description': description})

        if kwargs is not None:  # pragma: no cover
            update_dict.update(**kwargs)

        response = self._client._request('PUT',
                                         self._client._build_url('part2', part_id=self.id),
                                         params=API_EXTRA_PARAMS['part2'],
                                         json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Part ({})".format(response))

        if name:
            self.name = name

    def proxy_model(self) -> 'Part2':
        """
        Retrieve the proxy model of this proxied `Part` as a `Part`.

        Allows you to retrieve the model of a proxy. But trying to get the catalog model of a part that
        has no proxy, will raise an :exc:`NotFoundError`. Only models can have a proxy.

        .. versionchanged:: 3.0
           Added compatibility with KE-chain 3 backend.

        :return: :class:`Part` with category `MODEL` and from which the current part is proxied
        :raises NotFoundError: When no proxy model is found

        Example
        -------
        >>> proxy_part = project.model('Proxy based on catalog model')
        >>> catalog_model_of_proxy_part = proxy_part.proxy_model()

        >>> proxied_material_of_the_bolt_model = project.model('Bolt Material')
        >>> proxy_basis_for_the_material_model = proxied_material_of_the_bolt_model.proxy_model()

        """
        if self.category != Category.MODEL:
            raise IllegalArgumentError(
                'Part "{}" is not a product model, therefore it cannot have a proxy model'.format(self))
        if self.classification != Classification.PRODUCT or not self._proxy_model_id:
            raise NotFoundError('Part "{}" is not a product model, therefore it cannot have a proxy model'.format(self))

        return self._client.model(pk=self._proxy_model_id)

    def add(self, model: 'Part2', **kwargs) -> 'Part2':
        """Add a new child instance, based on a model, to this part.

        This can only act on instances. It needs a model from which to create the child instance.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param model: `Part2` object with category `MODEL`.
        :type model: :class:`Part2`
        :param kwargs: (optional) additional keyword=value arguments
        :return: :class:`Part2` with category `INSTANCE`.
        :raises APIError: if unable to add the new child instance

        Example
        -------
        >>> bike = project.part('Bike')
        >>> wheel_model = project.model('Wheel')
        >>> bike.add(wheel_model)

        """
        if self.category != Category.INSTANCE:
            raise APIError("Part should be of category INSTANCE")

        new_instance = self._client.create_part(self, model, **kwargs)

        if self._cached_children is not None:
            self._cached_children.append(new_instance)

        return new_instance

    def add_to(self, parent: 'Part2', **kwargs) -> 'Part2':
        """Add a new instance of this model to a part.

        This works if the current part is a model and an instance of this model is to be added
        to a part instances in the tree.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param parent: part to add the new instance to
        :type parent: :class:`Part2`
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :return: :class:`Part` with category `INSTANCE`
        :raises APIError: if unable to add the new child instance

        Example
        -------
        >>> wheel_model = project.model('wheel')
        >>> bike = project.part('Bike')
        >>> wheel_model.add_to(bike)

        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")

        new_instance = self._client.create_part(parent, self, **kwargs)

        if parent._cached_children is not None:
            parent._cached_children.append(new_instance)

        return new_instance

    def add_model(self, *args, **kwargs) -> 'Part2':
        """Add a new child model to this model.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :return: a :class:`Part2` of category `MODEL`
        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")

        new_model = self._client.create_model(self, *args, **kwargs)

        if self._cached_children is not None:
            self._cached_children.append(new_model)

        return new_model

    def add_proxy_to(self, parent: 'Part2',
                     name: Text,
                     multiplicity: Multiplicity = Multiplicity.ONE_MANY,
                     **kwargs) -> 'Part2':
        """Add this model as a proxy to another parent model.

        This will add the current model as a proxy model to another parent model. It ensure that it will copy the
        whole subassembly to the 'parent' model.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        :param name: Name of the new proxy model
        :type name: basestring
        :param parent: parent of the to be proxied model
        :type parent: :class:`Part`
        :param multiplicity: the multiplicity of the new proxy model (default ONE_MANY)
        :type multiplicity: basestring or None
        :param kwargs: (optional) additional kwargs that will be passed in the during the edit/update request
        :return: the new proxied :class:`Part`.
        :raises APIError: in case an Error occurs

        Examples
        --------
        >>> from pykechain.enums import Multiplicity
        >>> bike_model = project.model('Bike')
        # find the catalog model container, the highest parent to create catalog models under
        >>> catalog_model_container = project.model('Catalog container')
        >>> new_wheel_model = project.create_model(catalog_model_container, 'Wheel Catalog',
        ...                                        multiplicity=Multiplicity.ZERO_MANY)
        >>> new_wheel_model.add_proxy_to(bike_model, "Wheel", multiplicity=Multiplicity.ONE_MANY)

        """
        return self._client.create_proxy_model(self, parent, name, multiplicity, **kwargs)

    def add_property(self, *args, **kwargs) -> 'AnyProperty':
        """Add a new property to this model.

        See :class:`pykechain.Client.create_property` for available parameters.

        :return: :class:`Property`
        :raises APIError: in case an Error occurs
        """
        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")

        return self._client.create_property(self, *args, **kwargs)

    @staticmethod
    def _parse_update_dict(part: 'Part2',
                           properties_fvalues: List,
                           update_dict: Dict) -> Tuple[List[Dict], List[Dict]]:
        """
        Check the content of the update dict and insert them into the properties_fvalues list.

        :param part: Depending on whether you add to or update a part, this is the model or the part itself, resp.
        :param properties_fvalues: list of property values
        :param update_dict: dictionary with property values, keyed by property names
        :return: Tuple with 2 lists of dicts
        :rtype tuple
        """
        properties_fvalues = properties_fvalues or list()
        exception_fvalues = list()
        update_dict = update_dict or dict()

        if part.category == Category.INSTANCE:
            key = 'id'
        else:
            key = 'model_id'

        from pykechain.models import Base

        def make_serializable(value):
            # if the value is a reference property to another 'Base' Part, replace with its ID
            if isinstance(value, Base):
                return value.id
            else:
                return value

        from pykechain.models import AttachmentProperty2, DatetimeProperty2

        for prop_name_or_id, property_value in update_dict.items():
            property_to_update = part.property(prop_name_or_id)
            if isinstance(property_value, (list, set, tuple)):
                property_value = list(map(make_serializable, property_value))
            else:
                make_serializable(property_value)

            if isinstance(property_to_update, DatetimeProperty2) and isinstance(property_value, datetime.datetime):
                property_value = DatetimeProperty2.to_iso_format(property_value)

            updated_p = dict(
                value=property_value
            )
            if is_uuid(prop_name_or_id):
                updated_p[key] = prop_name_or_id
            else:
                updated_p[key] = property_to_update.id

            if isinstance(property_to_update, AttachmentProperty2):
                exception_fvalues.append(updated_p)
            else:
                properties_fvalues.append(updated_p)

        return properties_fvalues, exception_fvalues

    def add_with_properties(self,
                            model: 'Part2',
                            name: Optional[Text] = None,
                            update_dict: Optional[Dict] = None,
                            properties_fvalues: Optional[List[Dict]] = None,
                            refresh: Optional[bool] = None,
                            **kwargs) -> 'Part2':
        """
        Add a new part instance of a model as a child of this part instance and update its properties in one go.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        With KE-chain 3 backends you may now provide a whole set of properties to update using a `properties_fvalues`
        list of dicts. This will be merged with the `update_dict` optionally provided.
        The `properties_fvalues` list is a list of dicts containing at least the `id` and a `value`, but other keys
        may provided as well in the single update eg. `value_options`. Example::
            `properties_fvalues = [ {"id":<uuid of prop>, "value":<new_prop_value>}, {...}, ...]`

        .. versionchanged:: 3.0
           Added compatibility with KE-chain 3 backend. You may provide properties_fvalues as kwarg.
           Bulk option removed.

        .. versionchanged:: 3.3
           The 'refresh' flag is pending deprecation in version 3.4.. This flag had an effect of refreshing
           the list of children of the current part and was default set to True. This resulted in large
           processing times int he API as every `add_with_properties()` the children of the parent where all
           retrieved. The default is now 'False'. The part just created is however to the internal list of children
           once these children are retrieved earlier.

        :param model: model of the part which to add a new instance, should follow the model tree in KE-chain
        :type model: :class:`Part`
        :param name: (optional) name provided for the new instance as string otherwise use the name of the model
        :type name: basestring or None
        :param update_dict: dictionary with keys being property names (str) or property_id (from the property models)
                            and values being property values
        :type update_dict: dict or None
        :param properties_fvalues: (optional) keyword argument with raw list of properties update dicts
        :type properties_fvalues: list of dict or None
        :param refresh: (optional) refresh the children of this part after successful completion, default to False
        :type refresh: Optional[bool]
        :param kwargs: (optional) additional keyword arguments that will be passed inside the update request
        :return: the newly created :class:`Part`
        :raises NotFoundError: when the property name is not a valid property of this part
        :raises APIError: in case an Error occurs
        :raises IllegalArgumentError: in case of illegal arguments.

        Examples
        --------
        >>> bike = project.part('Bike')
        >>> wheel_model = project.model('Wheel')
        >>> bike.add_with_properties(wheel_model, 'Wooden Wheel', {'Spokes': 11, 'Material': 'Wood'})

        """
        if self.category != Category.INSTANCE:
            raise APIError("Part should be of category INSTANCE")

        if properties_fvalues and not isinstance(properties_fvalues, list):
            raise IllegalArgumentError("optional `properties_fvalues` need to be provided as a list of dicts")

        name = name or model.name
        url = self._client._build_url('parts2_new_instance')

        properties_fvalues, exception_fvalues = self._parse_update_dict(model, properties_fvalues, update_dict)

        response = self._client._request(
            'POST', url,
            params=API_EXTRA_PARAMS['parts2'],
            json=dict(
                name=name,
                model_id=model.id,
                parent_id=self.id,
                properties_fvalues=properties_fvalues,
                **kwargs
            )
        )

        if response.status_code != requests.codes.created:  # pragma: no cover
            raise APIError('{}: {}'.format(str(response), response.content))

        new_part_instance = Part2(response.json()['results'][0], client=self._client)

        # ensure that cached children are updated
        if self._cached_children is not None:
            self._cached_children.append(new_part_instance)
        if refresh is not None:
            import warnings
            warnings.warn("The refresh argument on the `add_with_properties()` method will be removed in "
                          "pykechain version 3.4.0 onwards. As the concept of 'refresh' had an effect based "
                          "on deprecated behaviour of the KE-chain API.", PendingDeprecationWarning)
            if refresh:
                # refreshed the children of this part (which is the parent of the children to be refreshed.
                self.children()

        # If any values were not set via the json, set them individually
        for exception_fvalue in exception_fvalues:
            property_model_id = exception_fvalue['model_id']
            property_instance = [p for p in new_part_instance.properties
                                 if p._json_data['model_id'] == property_model_id][0]
            property_instance.value = exception_fvalue['value']

        return new_part_instance

    def clone(self, **kwargs) -> 'Part2':
        """
        Clone a part.

        An optional name of the cloned part may be provided. If not provided the name will be set
        to "CLONE - <part name>". (available for KE-chain 3 backends only)
        An optional multiplicity, may be added as paremeter for the cloning of models. If not
        provided the multiplicity of the part will be used. (available for KE-chain 3 backends only)

        .. versionadded:: 2.3
        .. versionchanged:: 3.0
           Added optional paramenters the name and multiplicity for KE-chain 3 backends.

        :param kwargs: (optional) additional keyword=value arguments
        :return: cloned :class:`models.Part`
        :raises APIError: if the `Part` could not be cloned

        Example
        -------
        >>> bike = project.model('Bike')
        >>> bike2 = bike.clone()

        For KE-chain 3 backends

        >>> bike = project.model('Bike')
        >>> bike2 = bike.clone(name='Trike', multiplicity=Multiplicity.ZERO_MANY)

        """
        parent = self.parent()
        return self._client._create_clone(parent, self, **kwargs)

    def copy(self,
             target_parent: 'Part2',
             name: Optional[Text] = None,
             include_children: bool = True,
             include_instances: bool = True) -> 'Part2':
        """
        Copy the `Part` to target parent, both of them having the same category.

        .. versionadded:: 2.3

        :param target_parent: `Part` object under which the desired `Part` is copied
        :type target_parent: :class:`Part`
        :param name: how the copied top-level `Part` should be called
        :type name: basestring
        :param include_children: True to copy also the descendants of `Part`.
        :type include_children: bool
        :param include_instances: True to copy also the instances of `Part` to ALL the instances of target_parent.
        :type include_instances: bool
        :returns: copied :class:`Part` model.
        :raises IllegalArgumentError: if part and target_parent have different `Category`
        :raises IllegalArgumentError: if part and target_parent are identical

        Example
        -------
        >>> model_to_copy = project.model(name='Model to be copied')
        >>> bike = project.model('Bike')
        >>> model_to_copy.copy(target_parent=bike, name='Copied model',
        >>>                    include_children=True,
        >>>                    include_instances=True)

        """
        get_mapping_dictionary(clean=True)
        get_edited_one_many(clean=True)

        # to ensure that all properties are retrieved from the backend
        # as it might be the case that a part is retrieved in the context of a widget and there could be a possibility
        # that not all properties are retrieved we perform a refresh of the part itself first.
        self.refresh()

        if not isinstance(target_parent, Part2):
            raise IllegalArgumentError("`target_parent` needs to be a part, got '{}'".format(type(target_parent)))

        if self.category == Category.MODEL and target_parent.category == Category.MODEL:
            # Cannot add a model under an instance or vice versa
            copied_model = relocate_model(part=self, target_parent=target_parent, name=name,
                                          include_children=include_children)
            if include_instances:
                instances_to_be_copied = list(self.instances())
                parent_instances = list(target_parent.instances())
                for parent_instance in parent_instances:
                    for instance in instances_to_be_copied:
                        instance.populate_descendants()
                        move_part_instance(part_instance=instance, target_parent=parent_instance,
                                           part_model=self, name=instance.name,
                                           include_children=include_children)
            return copied_model

        elif self.category == Category.INSTANCE and target_parent.category == Category.INSTANCE:
            copied_instance = relocate_instance(part=self, target_parent=target_parent, name=name,
                                                include_children=include_children)
            return copied_instance
        else:
            raise IllegalArgumentError('part "{}" and target parent "{}" must have the same category'.
                                       format(self.name, target_parent.name))

    def move(self,
             target_parent: 'Part2',
             name: Optional[Text] = None,
             include_children: bool = True,
             include_instances: bool = True) -> 'Part2':
        """
        Move the `Part` to target parent, both of them the same category.

        .. versionadded:: 2.3

        :param target_parent: `Part` object under which the desired `Part` is moved
        :type target_parent: :class:`Part`
        :param name: how the moved top-level `Part` should be called
        :type name: basestring
        :param include_children: True to move also the descendants of `Part`. If False, the children will be lost.
        :type include_children: bool
        :param include_instances: True to move also the instances of `Part` to ALL the instances of target_parent.
        :type include_instances: bool
        :returns: moved :class:`Part` model.
        :raises IllegalArgumentError: if part and target_parent have different `Category`
        :raises IllegalArgumentError: if target_parent is descendant of part

        Example
        -------
        >>> model_to_move = project.model(name='Model to be moved')
        >>> bike = project.model('Bike')
        >>> model_to_move.move(target_parent=bike, name='Moved model',
        >>>                    include_children=True,
        >>>                    include_instances=True)

        """
        self.refresh()
        get_mapping_dictionary(clean=True)
        get_edited_one_many(clean=True)

        if not name:
            name = self.name
        if self.category == Category.MODEL and target_parent.category == Category.MODEL:
            moved_model = relocate_model(part=self, target_parent=target_parent, name=name,
                                         include_children=include_children)
            if include_instances:
                retrieve_instances_to_copied = list(self.instances())
                retrieve_parent_instances = list(target_parent.instances())
                for parent_instance in retrieve_parent_instances:
                    for instance in retrieve_instances_to_copied:
                        instance.populate_descendants()
                        move_part_instance(part_instance=instance, target_parent=parent_instance,
                                           part_model=self, name=instance.name, include_children=include_children)
            self.delete()
            return moved_model
        elif self.category == Category.INSTANCE and target_parent.category == Category.INSTANCE:
            moved_instance = relocate_instance(part=self, target_parent=target_parent, name=name,
                                               include_children=include_children)
            try:
                self.delete()
            except APIError:
                model_of_instance = self.model()
                model_of_instance.delete()
            return moved_instance
        else:
            raise IllegalArgumentError('part "{}" and target parent "{}" must have the same category')

    def update(self,
               name: Optional[Text] = None,
               update_dict: Optional[Dict] = None,
               properties_fvalues: Optional[List[Dict]] = None,
               **kwargs) -> None:
        """
        Edit part name and property values in one go.

        In order to prevent the backend from updating the frontend you may add `suppress_kevents=True` as
        additional keyword=value argument to this method. This will improve performance of the backend
        against a trade-off that someone looking at the frontend won't notice any changes unless the page
        is refreshed.

        With KE-chain 3 backends you may now provide a whole set of properties to update using a `properties_fvalues`
        list of dicts. This will be merged with the `update_dict` optionally provided.
        The `properties_fvalues` list is a list of dicts containing at least the `id` and a `value`, but other keys
        may provided as well in the single update eg. `value_options`. Example::
            `properties_fvalues = [ {"id":<uuid of prop>, "value":<new_prop_value>}, {...}, ...]`

        .. versionchanged:: 3.0
           Added compatibility with KE-chain 3 backend. You may provide properties_fvalues as kwarg.
           Bulk option removed.

        :param name: new part name (defined as a string)
        :type name: basestring or None
        :param update_dict: dictionary with keys being property names (str) or property ids (uuid)
                            and values being property values
        :type update_dict: dict or None
        :param properties_fvalues: (optional) keyword argument with raw list of properties update dicts
        :type properties_fvalues: list of dict or None
        :param kwargs: additional keyword-value arguments that will be passed into the part update request.
        :return: the updated :class:`Part`
        :raises NotFoundError: when the property name is not a valid property of this part
        :raises IllegalArgumentError: when the type or value of an argument provided is incorrect
        :raises APIError: in case an Error occurs

        Example
        -------
        >>> bike = project.part('Bike')
        >>> bike.update(name='Good name', update_dict={'Gears': 11, 'Total Height': 56.3})

        Example with properties_fvalues: <pyke Part2 'Copied model under Bike' id 95d35be6>

        >>> bike = project.part('Bike')
        >>> bike.update(name='Good name',
        ...             properties_fvalues=[{'id': '95d35be6...', 'value': 11},
        ...                                 {'id': '7893cba4...', 'value': 56.3, 'value_options': {...}})

        """
        # dict(name=name, properties=json.dumps(update_dict))) with property ids:value
        # action = 'bulk_update_properties'  # not for KEC3
        if name and not isinstance(name, (string_types, text_type)):
            raise IllegalArgumentError("Name of the part should be provided as a string")

        if properties_fvalues and not isinstance(properties_fvalues, list):
            raise IllegalArgumentError("optional `properties_fvalues` need to be provided as a list of dicts")

        properties_fvalues, exception_fvalues = self._parse_update_dict(self, properties_fvalues, update_dict)

        payload_json = dict(
            properties_fvalues=properties_fvalues,
            **kwargs
        )

        if name:
            payload_json.update(name=name)

        response = self._client._request(
            'PUT', self._client._build_url('part2', part_id=self.id),
            params=API_EXTRA_PARAMS['part2'],
            json=payload_json
        )

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update the part '{}', got: '{}'".format(self, response.content))

        # update local properties (without a call)
        self.refresh(json=response.json()['results'][0])

        # If any values were not set via the json, set them individually
        for exception_fvalue in exception_fvalues:
            self.property(exception_fvalue['id']).value = exception_fvalue['value']

    def delete(self) -> None:
        """Delete this part.

        :return: None
        :raises APIError: in case an Error occurs
        """
        response = self._client._request('DELETE', self._client._build_url('part2', part_id=self.id))

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            raise APIError("Could not delete part: {} with id {}: ({})".format(self.name, self.id, response))

    def order_properties(self, property_list: Optional[List[Text]] = None) -> None:
        """
        Order the properties of a part model using a list of property objects or property names or property id's.

        For KE-chain 3 backends, the order can also directly provided as a unique integer for each property in the
        `Part.update()` function if you provide the `properties_fvalues` list of dicts yourself. For more information
        please refer to the `Part2.update()` documentation.

        .. versionchanged:: 3.0
           For KE-chain 3 backend the `Part.update()` method is used with a properties_fvalues list of dicts.

        :param property_list: ordered list of property names (basestring) or property id's (uuid) or
                              a `Property` object.
        :type property_list: list(basestring)
        :returns: the :class:`Part` with the reordered list of properties
        :raises APIError: when an Error occurs
        :raises IllegalArgumentError: When provided a wrong argument

        Examples
        --------
        >>> front_fork = project.model('Front Fork')
        >>> front_fork.order_properties(['Material', 'Height (mm)', 'Color'])

        >>> front_fork = project.model('Front Fork')
        >>> material = front_fork.property('Material')
        >>> height = front_fork.property('Height (mm)')
        >>> color = front_fork.property('Color')
        >>> front_fork.order_properties([material, height, color])

        Alternatively you may use the `Part.update()` function to directly alter the order of the properties and
        eventually even more (defaut) model values.

        >>> front_fork.update(properties_fvalues= [{'id': material.id, 'order':10},
        ...                                        {'id': height.id, 'order': 20, 'value':13.37}
        ...                                        {'id': color.id, 'order':  30, 'name': 'Colour'}])

        """
        # in KEC3 backend we can (re)use the part update endpoint with a properties_fvalues list of dicts
        # properties_fvalues = [ {'model_id':<uuid>, 'order': <int> ]

        if self.category != Category.MODEL:
            raise APIError("Part should be of category MODEL")
        if not isinstance(property_list, list):
            raise IllegalArgumentError('Expected a list of strings or Property() objects, got a {} object'.
                                       format(type(property_list)))

        properties_fvalues = list()
        for order, prop_name_or_id in enumerate(property_list):
            updated_p = {'order': order}
            if is_uuid(prop_name_or_id):
                updated_p['id'] = prop_name_or_id
            elif isinstance(prop_name_or_id, Property2):
                updated_p['id'] = prop_name_or_id.id
            else:
                updated_p['id'] = self.property(prop_name_or_id).id
            properties_fvalues.append(updated_p)

        return self.update(properties_fvalues=properties_fvalues)

    #
    # Utility Functions
    #

    def _repr_html_(self) -> Text:
        """
        Represent the part in a HTML table for the use in notebooks.

        :return: html text
        :rtype: Text
        """
        html = [
            "<table width=100%>",
            "<caption>{}</caption>".format(self.name),
            "<tr>",
            "<th>Property</th>",
            "<th>Value</th>",
            "</tr>"
        ]

        for prop in self.properties:
            style = "color:blue;" if prop._json_data.get('output', False) else ""

            html.append("<tr style=\"{}\">".format(style))
            html.append("<td>{}</td>".format(prop.name))
            html.append("<td>{}</td>".format(prop.value))
            html.append("</tr>")

        html.append("</table>")

        return ''.join(html)

    def as_dict(self) -> Dict:
        """
        Retrieve the properties of a part inside a dict in this structure: {property_name: property_value}.

        .. versionadded:: 1.9

        :returns: the values of the properties as a `dict`
        :rtype: dict

        Example
        -------
        >>> front_wheel = project.part('Front Wheel')
        >>> front_wheel_properties = front_wheel.as_dict()
        {'Diameter': 60.8,
         'Spokes': 24,
         'Rim Material': 'Aluminium',
         'Tire Thickness': 4.2}

        """
        properties_dict = dict()
        for prop in self.properties:
            properties_dict[prop.name] = prop.value
        return properties_dict



from datetime import datetime

from pykechain.enums import Multiplicity, PropertyType
from pykechain.exceptions import NotFoundError, IllegalArgumentError
from tests.classes import TestBetamax


class TestPartsCopyMove(TestBetamax):

    def setUp(self):
        super(TestPartsCopyMove, self).setUp()
        self.base = self.project.model(name__startswith='Catalog')
        self.cross_scope_project = self.client.scope(ref='cannondale-project')
        self.cross_scope_bike = self.cross_scope_project.model(ref='cannondale-bike')
        self.cross_scope_wheel = self.cross_scope_bike.add_model(name='__Wheel')

        self.model_to_be_copied = self.project.create_model(
            parent=self.base, name='__Model to be Copied @ {} [TEST]'.format(str(datetime.now())),
            multiplicity=Multiplicity.ONE_MANY)  # type Part
        p_char = self.model_to_be_copied.add_property(
            name='__Property single text',
            description='Description of Property single text',
            unit='mm',
            property_type=PropertyType.CHAR_VALUE
        )
        p_decimal = self.model_to_be_copied.add_property(
            name='__Property decimal number',
            default_value=33,
            property_type=PropertyType.FLOAT_VALUE
        )
        p_single_select = self.model_to_be_copied.add_property(
            name='__Property single select list',
            options=dict(value_choices=['a', 'b', 'c']),
            property_type=PropertyType.SINGLE_SELECT_VALUE
        )
        p_multi_reference_property = self.model_to_be_copied.add_property(
            name='__Property multi reference',
            default_value=self.cross_scope_wheel,
            # options={S},
            property_type=PropertyType.REFERENCES_VALUE
        )
        sub_part1 = self.model_to_be_copied.add_model(
            name="__subpart 1",
            multiplicity=Multiplicity.ONE
        )
        sub_part2 = self.model_to_be_copied.add_model(
            name="__subpart 2",
            multiplicity=Multiplicity.ONE
        )

        self.model_target_parent = self.project.model(ref='bike')
        self.instance_to_be_copied = self.model_to_be_copied.instances()[0]
        self.instance_target_parent = self.model_to_be_copied.parent().instances()[0]
        self.dump_part = None

    def tearDown(self):
        self.model_to_be_copied.delete()
        self.cross_scope_wheel.delete()
        if self.dump_part:
            self.dump_part.delete()
        super(TestPartsCopyMove, self).tearDown()

    def test_copy_part_model_given_name_include_children(self):
        model_target_parent = self.project.model('Bike')
        copied_model = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name='__Copied model under Bike',
            include_children=True,
            include_instances=False
        )
        copied_model.populate_descendants()
        self.dump_part = copied_model

        # testing
        self.assertTrue(copied_model)
        self.assertEqual(copied_model.property('__Property single text')._json_data['description'],
                         'Description of Property single text')
        self.assertEqual(copied_model.property('__Property single text')._json_data['unit'], 'mm')
        self.assertEqual(copied_model.property('__Property decimal number').value, 33)
        self.assertEqual(copied_model.property('__Property single select list').options, ['a', 'b', 'c'])
        self.assertEqual(len(copied_model._cached_children), 2)

    def test_copy_part_model_include_instances(self):
        model_target_parent = self.project.model('Bike')
        self.dump_part = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name='__Copied model under Bike',
            include_children=True,
            include_instances=True
        )
        bike_part = self.project.part('Bike')

        copied_instance = [child for child in bike_part.children() if 'to be Copied' in child.name][0]
        copied_instance.populate_descendants()

        # testing
        self.assertTrue(copied_instance)
        self.assertEqual(len(copied_instance.properties), 4)
        self.assertTrue(copied_instance.property('__property-single-text'))
        self.assertEqual(copied_instance.property('__property-decimal-number').value, 33)
        self.assertEqual(copied_instance.property('__property-single-select-list').value, None)

        self.assertEqual(len(copied_instance._cached_children), 2)

    def test_copy_part_model_empty_name_not_include_children(self):
        # setUp
        model_target_parent = self.project.model('Bike')
        name_of_part = '__Copied model under Bike'
        copied_model = self.model_to_be_copied.copy(
            target_parent=model_target_parent,
            name=name_of_part,
            include_children=False,
            include_instances=False
        )

        copied_model.populate_descendants()
        self.dump_part = copied_model

        # testing
        self.assertTrue(copied_model)
        self.assertEqual(copied_model.name, name_of_part)
        self.assertEqual(len(copied_model.properties), 4)
        self.assertEqual(len(copied_model._cached_children), 0)

    def test_move_part_model(self):
        # setUp
        model_target_parent = self.project.model('Bike')
        clone_of_original_model = self.model_to_be_copied.clone()
        self.assertEqual(clone_of_original_model.name, 'CLONE - {}'.format(self.model_to_be_copied.name))

        clone_of_original_model.move(target_parent=model_target_parent,
                                     name='__New part under bike',
                                     include_children=True)

        self.dump_part = self.project.model(name='__New part under bike')

        # testing
        with self.assertRaises(NotFoundError):
            self.project.model(name=clone_of_original_model.name)

    def test_copy_part_instance(self):
        # setUp
        instance_to_be_copied = self.model_to_be_copied.instances()[0]
        instance_target_parent = self.project.part('Bike')
        instance_to_be_copied.copy(target_parent=instance_target_parent, name='__Copied instance', include_children=True)

        copied_instance = self.project.part(name='__Copied instance')
        copied_instance.populate_descendants()
        self.dump_part = copied_instance.model()

        # testing
        self.assertTrue(copied_instance)
        self.assertEqual(copied_instance.name, '__Copied instance')
        self.assertEqual(len(copied_instance.properties), len(self.model_to_be_copied.properties))
        self.assertEqual(copied_instance.property('__property-single-text').value, None)
        self.assertEqual(copied_instance.property('__property-decimal-number').value, 33)
        self.assertEqual(copied_instance.property('__property-single-select-list').value, None)

        self.assertEqual(len(copied_instance._cached_children), 2)

    def test_move_part_instance(self):
        # setUp
        multiplicity_model = self.model_to_be_copied.add_model(
            name="__multiplicity part",
            multiplicity=Multiplicity.ZERO_MANY
        )
        multiplicity_part1 = self.instance_to_be_copied.add(
            model=multiplicity_model,
            name="__multiplicity part instance 1",
        )
        multiplicity_part2 = self.instance_to_be_copied.add(
            model=multiplicity_model,
            name="__multiplicity part instance 2",
        )

        clone_of_original_instance = multiplicity_part2.clone()

        moved_instance = clone_of_original_instance.move(target_parent=self.project.part('Bike'),
                                                         name='__Moved clone instance',
                                                         include_children=False)

        moved_instance.populate_descendants()
        self.dump_part = moved_instance.model()

        # testing
        with self.assertRaises(NotFoundError):
            self.project.model(name=clone_of_original_instance.name)

        self.assertTrue(moved_instance)
        self.assertEqual(len(moved_instance.properties), 0)
        # self.assertEqual(moved_instance.property('Property multi-text').value, 'Yes yes oh yes')

    def test_copy_different_categories(self):
        target_instance = self.project.part(name='Bike')

        with self.assertRaisesRegex(IllegalArgumentError, 'must have the same category'):
            self.model_to_be_copied.copy(target_parent=target_instance)

    def test_move_different_categories(self):
        target_instance = self.project.part(name='Bike')

        with self.assertRaisesRegex(IllegalArgumentError, 'must have the same category'):
            self.model_to_be_copied.move(target_parent=target_instance)



from pykechain.enums import PropertyType, FilterType, Multiplicity
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import MultiReferenceProperty2
from pykechain.models.validators import RequiredFieldValidator
from tests.classes import TestBetamax


class TestMultiReferenceProperty(TestBetamax):

    def setUp(self):
        super(TestMultiReferenceProperty, self).setUp()

        # reference Part target (model and 1 instance)
        _wheel_model = self.project.model(ref='wheel')

        props = [
            PropertyType.DATETIME_VALUE,
            PropertyType.SINGLE_SELECT_VALUE,
            PropertyType.BOOLEAN_VALUE,
            PropertyType.FLOAT_VALUE,
            PropertyType.INT_VALUE,
            PropertyType.CHAR_VALUE,
        ]

        self.target_model = self.project.create_model_with_properties(
            parent=_wheel_model.parent_id,
            name="__Wheel",
            multiplicity=Multiplicity.ONE_MANY,
            properties_fvalues=[dict(name=p, property_type=p) for p in props],
        )

        # reference property model (with a value pointing to a reference target part model
        self.part_model = self.project.model('Bike')

        self.datetime_prop = self.target_model.property(PropertyType.DATETIME_VALUE)
        self.ssl_prop = self.target_model.property(PropertyType.SINGLE_SELECT_VALUE)
        self.bool_prop = self.target_model.property(PropertyType.BOOLEAN_VALUE)
        self.float_prop = self.target_model.property(PropertyType.FLOAT_VALUE)
        self.integer_prop = self.target_model.property(PropertyType.INT_VALUE)
        self.char_prop = self.target_model.property(PropertyType.CHAR_VALUE)

        self.ref_prop_name = '__Test reference property'
        self.ref_prop_model = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.target_model.id
        )  # type: MultiReferenceProperty2

        # reference property instance holding the value
        self.ref = self.part_model.instance().property(self.ref_prop_name)

    def tearDown(self):
        self.target_model.delete()
        self.ref_prop_model.delete()
        super(TestMultiReferenceProperty, self).tearDown()

    def test_referencing_a_model(self):
        # setUp
        wheel_model = self.project.model('Wheel')
        self.ref_prop_model.value = [wheel_model]

        # testing
        self.assertEqual(len(self.ref_prop_model.value), 1)

    def test_referencing_multiple_instances_using_parts(self):
        # setUp
        wheel_model = self.project.model('Wheel')
        self.ref_prop_model.value = [wheel_model]
        wheel_instances = wheel_model.instances()
        wheel_instances_list = [instance for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list

        # testing
        self.assertTrue(len(self.ref.value) >= 2)

    def test_referencing_multiple_instances_using_ids(self):
        # setUp
        wheel_model = self.project.model('Wheel')
        self.ref_prop_model.value = [wheel_model]
        wheel_instances = wheel_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list
        self.ref._cached_values = None

        # testing
        self.assertTrue(len(self.ref.value) >= 2)

    def test_referencing_a_part_not_in_a_list(self):
        # setUp
        front_wheel = self.project.part('Front Wheel')

        # testing
        with self.assertRaises(ValueError):
            self.ref.value = front_wheel

    def test_referencing_a_list_with_no_parts(self):
        # setUp
        fake_part = [15, 21, 26]

        # testing
        with self.assertRaises(ValueError):
            self.ref.value = fake_part

    def test_value_if_multi_ref_gives_back_all_parts(self):
        """because of #276 problem"""
        # setUp
        self.ref_prop_model.value = [self.target_model]

        wheel_instances = self.target_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list
        self.ref.refresh()

        # testing
        all_referred_parts = self.ref.value
        self.assertEqual(len(all_referred_parts), len(self.ref._value))

    def test_value_if_nothing_is_referenced(self):
        # setUp
        value_of_multi_ref = self.ref.value

        # testing
        self.assertFalse(self.ref.has_value())
        self.assertIsNone(value_of_multi_ref)

    def test_multi_ref_choices(self):
        # setUp
        self.project.part(ref='bike').add(model=self.target_model, name='__Wheel 2')
        self.ref_prop_model.value = [self.target_model]
        possible_options = self.ref.choices()

        # testing
        self.assertEqual(2, len(possible_options))

    def test_create_ref_property_referencing_part_in_list(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=[self.target_model],
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_referencing_id_in_list(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=[self.target_model.id]
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_wrongly_referencing_in_list(self):
        # testing
        with self.assertRaises(IllegalArgumentError):
            self.part_model.add_property(
                name=self.ref_prop_name,
                property_type=PropertyType.REFERENCES_VALUE,
                default_value=[12]
            )

    def test_create_ref_property_referencing_part(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.target_model
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_referencing_id(self):
        # setUp
        new_reference_to_wheel = self.part_model.add_property(
            name=self.ref_prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
            default_value=self.target_model.id
        )
        # testing
        self.assertTrue(self.ref_prop_model.value[0].id, self.target_model.id)

        # tearDown
        new_reference_to_wheel.delete()

    def test_create_ref_property_wrongly_referencing(self):
        # testing
        with self.assertRaises(IllegalArgumentError):
            self.part_model.add_property(
                name=self.ref_prop_name,
                property_type=PropertyType.REFERENCES_VALUE,
                default_value=True
            )

    # new in 3.0
    def test_set_prefilters_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop  # decimal property
        spokes_property = self.integer_prop  # integer property
        rim_material_property = self.char_prop  # single line text
        import datetime
        new_year_2020 = datetime.datetime(year=2020, month=1, day=1)

        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property,
                             spokes_property,
                             rim_material_property,
                             self.datetime_prop,
                             self.ssl_prop,
                             self.bool_prop],
            values=[30.5,
                    7,
                    'Al',
                    new_year_2020,
                    'Michelin',
                    True],
            filters_type=[FilterType.GREATER_THAN_EQUAL,
                          FilterType.LOWER_THAN_EQUAL,
                          FilterType.CONTAINS,
                          FilterType.GREATER_THAN_EQUAL,
                          FilterType.CONTAINS,
                          FilterType.EXACT]
        )
        self.assertIn('property_value', self.ref_prop_model._options['prefilters'])

        filter_string = self.ref_prop_model._options['prefilters']['property_value']
        filters = set(filter_string.split(','))

        # testing
        self.assertIn("{}:{}:{}".format(diameter_property.id, 30.5, FilterType.GREATER_THAN_EQUAL), filters)
        self.assertIn("{}:{}:{}".format(spokes_property.id, 7, FilterType.LOWER_THAN_EQUAL), filters)
        self.assertIn("{}:{}:{}".format(rim_material_property.id, 'Al', FilterType.CONTAINS), filters)
        self.assertIn("{}:{}:{}".format(self.bool_prop.id, 'true', FilterType.EXACT), filters)
        self.assertIn("{}:{}:{}".format(self.ssl_prop.id, 'Michelin', FilterType.CONTAINS), filters)
        self.assertIn("{}:{}:{}".format(self.datetime_prop.id, new_year_2020, FilterType.GREATER_THAN_EQUAL),
                      filters)

    def test_set_prefilters_on_reference_property_with_excluded_propmodels_and_validators(self):
        # The excluded propmodels and validators already set on the property should not be erased when
        # setting prefilters

        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_excluded_propmodels(property_models=[diameter_property, spokes_property])
        self.ref_prop_model.validators = [RequiredFieldValidator()]

        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL]
        )

        # testing Filters
        self.assertIn("{}:{}:{}".format(diameter_property.id, 15.13, FilterType.GREATER_THAN_EQUAL),
                      self.ref_prop_model._options['prefilters']['property_value'])

        # testing Excluded props
        self.assertIn('propmodels_excl', self.ref_prop_model._options)
        excluded = self.ref_prop_model._options['propmodels_excl']

        self.assertEqual(len(excluded), 2)
        self.assertIn(diameter_property.id, excluded)
        self.assertIn(spokes_property.id, excluded)

        # testing Validators
        self.assertTrue(self.ref_prop_model.validators)
        self.assertIsInstance(self.ref_prop_model._validators[0], RequiredFieldValidator)

    def test_add_prefilters_to_reference_property(self):
        # Set the initial prefilters
        self.ref_prop_model.set_prefilters(
            property_models=[self.float_prop],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL]
        )
        # Add others, checks whether the initial ones are remembered (overwrite = False)
        self.ref_prop_model.set_prefilters(
            property_models=[self.integer_prop],
            values=[2],
            filters_type=[FilterType.LOWER_THAN_EQUAL],
            overwrite=False
        )

        # testing
        self.assertTrue('property_value' in self.ref_prop_model._options['prefilters'])
        self.assertTrue("{}:{}:{}".format(self.float_prop.id, 15.13, FilterType.GREATER_THAN_EQUAL) in
                        self.ref_prop_model._options['prefilters']['property_value'])
        self.assertTrue("{}:{}:{}".format(self.integer_prop.id, 2, FilterType.LOWER_THAN_EQUAL) in
                        self.ref_prop_model._options['prefilters']['property_value'])

    def test_overwrite_prefilters_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        # Set the initial prefilters
        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL]
        )
        # Add others, see if the first ones are remembered (overwrite = False)
        self.ref_prop_model.set_prefilters(
            property_models=[spokes_property],
            values=[2],
            filters_type=[FilterType.LOWER_THAN_EQUAL],
            overwrite=True
        )

        # testing
        self.assertTrue('property_value' in self.ref_prop_model._options['prefilters'])
        self.assertTrue("{}:{}:{}".format(spokes_property.id, 2, FilterType.LOWER_THAN_EQUAL) ==
                        self.ref_prop_model._options['prefilters']['property_value'])

    def test_set_prefilters_on_reference_property_using_uuid(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property.id,
                             spokes_property.id],
            values=[30.5,
                    7],
            filters_type=[FilterType.GREATER_THAN_EQUAL,
                          FilterType.LOWER_THAN_EQUAL]
        )

        # testing
        self.assertTrue('property_value' in self.ref_prop_model._options['prefilters'])
        self.assertTrue("{}:{}:{}".format(diameter_property.id, 30.5, FilterType.GREATER_THAN_EQUAL) in
                        self.ref_prop_model._options['prefilters']['property_value'])
        self.assertTrue("{}:{}:{}".format(spokes_property.id, 7, FilterType.LOWER_THAN_EQUAL) in
                        self.ref_prop_model._options['prefilters']['property_value'])

    def test_set_prefilters_on_reference_property_the_wrong_way(self):
        # setUp
        bike_gears_property = self.part_model.property(name='Gears')
        instance_diameter_property = self.target_model.instances()[0].property(PropertyType.FLOAT_VALUE)
        diameter_property = self.float_prop

        # testing
        # When prefilters are being set, but the property does not belong to the referenced part
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(
                property_models=[bike_gears_property],
                values=[2],
                filters_type=[FilterType.GREATER_THAN_EQUAL]
            )

        # When prefilters are being set, but the property is an instance, not a model
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(
                property_models=[instance_diameter_property],
                values=[3.33],
                filters_type=[FilterType.GREATER_THAN_EQUAL]
            )

        # When prefilters are being set, but the size of lists is not consistent
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_prefilters(
                property_models=[diameter_property],
                values=[3.33, 1.51],
                filters_type=[FilterType.GREATER_THAN_EQUAL, FilterType.CONTAINS]
            )

        # When prefilters are being set, but no UUIDs or `Property` objects are being used in property_models
        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.ref_prop_model.set_prefilters(
                property_models=[False, 301],
                values=[3.33, 1.51],
                filters_type=[FilterType.GREATER_THAN_EQUAL, FilterType.CONTAINS]
            )

    def test_set_excluded_propmodels_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_excluded_propmodels(property_models=[diameter_property, spokes_property])

        # testing
        self.assertEqual(len(self.ref_prop_model._options['propmodels_excl']), 2)
        self.assertTrue(diameter_property.id in self.ref_prop_model._options['propmodels_excl'])
        self.assertTrue(spokes_property.id in self.ref_prop_model._options['propmodels_excl'])

    def test_set_excluded_propmodels_on_reference_property_using_uuid(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        self.ref_prop_model.set_excluded_propmodels(property_models=[diameter_property.id, spokes_property.id])

        # testing
        self.assertEqual(len(self.ref_prop_model._options['propmodels_excl']), 2)
        self.assertTrue(diameter_property.id in self.ref_prop_model._options['propmodels_excl'])
        self.assertTrue(spokes_property.id in self.ref_prop_model._options['propmodels_excl'])

    def test_set_excluded_propmodels_on_reference_property_with_prefilters_and_validators(self):
        # The prefilters and validators already set on the property should not be erased when setting
        # excluded propmodels

        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop
        self.ref_prop_model.set_prefilters(
            property_models=[diameter_property],
            values=[15.13],
            filters_type=[FilterType.GREATER_THAN_EQUAL]
        )

        self.ref_prop_model.validators = [RequiredFieldValidator()]

        self.ref_prop_model.set_excluded_propmodels(property_models=[diameter_property, spokes_property])

        # testing
        self.assertTrue('property_value' in self.ref_prop_model._options['prefilters'])
        self.assertTrue("{}:{}:{}".format(diameter_property.id, 15.13, FilterType.GREATER_THAN_EQUAL) in
                        self.ref_prop_model._options['prefilters']['property_value'])
        self.assertEqual(len(self.ref_prop_model._options['propmodels_excl']), 2)
        self.assertTrue(diameter_property.id in self.ref_prop_model._options['propmodels_excl'])
        self.assertTrue(spokes_property.id in self.ref_prop_model._options['propmodels_excl'])
        self.assertTrue(isinstance(self.ref_prop_model._validators[0], RequiredFieldValidator))

    def test_set_excluded_propmodels_on_reference_property_the_wrong_way(self):
        # setUp
        bike_gears_property = self.part_model.property(name='Gears')
        instance_diameter_property = self.target_model.instances()[0].property(PropertyType.FLOAT_VALUE)

        # testing
        # When excluded propmodels are being set, but the property does not belong to the referenced part
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_excluded_propmodels(property_models=[bike_gears_property])

        # When excluded propmodels are being set, but the property is an instance, not a model
        with self.assertRaises(IllegalArgumentError):
            self.ref_prop_model.set_excluded_propmodels(property_models=[instance_diameter_property])

        # When excluded propmodels are being set, but no UUIDs or `Property` objects are being used in property_models
        with self.assertRaises(IllegalArgumentError):
            # noinspection PyTypeChecker
            self.ref_prop_model.set_excluded_propmodels(property_models=[False, 301])

    def test_add_excluded_propmodels_to_reference_property(self):
        # Set the initial excluded propmodels
        self.ref_prop_model.set_excluded_propmodels(property_models=[self.float_prop])

        # Add others, checks whether the initial ones are remembered (overwrite = False)
        self.ref_prop_model.set_excluded_propmodels(property_models=[self.integer_prop], overwrite=False)

        # testing
        self.assertEqual(len(self.ref_prop_model._options['propmodels_excl']), 2)
        self.assertTrue(self.float_prop.id in self.ref_prop_model._options['propmodels_excl'])
        self.assertTrue(self.integer_prop.id in self.ref_prop_model._options['propmodels_excl'])

    def test_overwrite_excluded_propmodels_on_reference_property(self):
        # setUp
        diameter_property = self.float_prop
        spokes_property = self.integer_prop

        # Set the initial excluded propmodels
        self.ref_prop_model.set_excluded_propmodels(property_models=[diameter_property])

        # Overwrite them
        self.ref_prop_model.set_excluded_propmodels(property_models=[spokes_property], overwrite=True)

        # testing
        self.assertEqual(len(self.ref_prop_model._options['propmodels_excl']), 1)
        self.assertTrue(diameter_property.id not in self.ref_prop_model._options['propmodels_excl'])
        self.assertTrue(spokes_property.id in self.ref_prop_model._options['propmodels_excl'])

    def test_retrieve_scope_id(self):
        frame = self.project.part(name='Frame')

        ref_to_wheel = frame.property(name='Ref to wheel')

        self.assertEqual(ref_to_wheel.scope_id, self.project.id)

    def test_property_clear_selected_part_model(self):
        # setUp
        wheel_model = self.project.model('Wheel')
        self.ref_prop_model.value = [wheel_model]

        self.assertTrue(wheel_model.id == self.ref_prop_model.value[0].id)

        self.ref_prop_model.value = None

        # testing
        self.assertIsNone(self.ref_prop_model.value)

    def test_property_clear_referenced_part_instances(self):
        # setUp
        wheel_model = self.project.model('Wheel')
        self.ref_prop_model.value = [wheel_model]

        wheel_instances = wheel_model.instances()
        wheel_instances_list = [instance.id for instance in wheel_instances]

        # set ref value
        self.ref.value = wheel_instances_list

        # clear referenced part instances
        self.ref.value = None

        # testing
        self.assertIsNone(self.ref.value)
        self.assertIsNotNone(self.ref_prop_model.value)

    def test_copy_reference_property_with_options(self):
        # setUp
        copied_ref_property = self.ref_prop_model.copy(target_part=self.part_model,
                                                       name='__Copied ref property')

        # testing
        self.assertEqual(copied_ref_property.name, '__Copied ref property')
        self.assertEqual(copied_ref_property.description, self.ref_prop_model.description)
        self.assertEqual(copied_ref_property.unit, self.ref_prop_model.unit)
        self.assertEqual(copied_ref_property.value, self.ref_prop_model.value)
        self.assertDictEqual(copied_ref_property._options, self.ref_prop_model._options)

        # tearDown
        copied_ref_property.delete()


class TestMultiReferencePropertyXScope(TestBetamax):

    def setUp(self):
        super(TestMultiReferencePropertyXScope, self).setUp()
        self.x_scope = self.client.create_scope(name='Cross_reference scope', tags=['x-scope-target'])

        self.part_model = self.project.model('Bike')

        # Create reference property and retrieve its instance
        prop_name = 'cross-scope reference property'
        self.x_reference_model = self.part_model.add_property(
            name=prop_name,
            property_type=PropertyType.REFERENCES_VALUE,
        )
        self.x_reference = self.part_model.instance().property(prop_name)

        # Define target model and create an instance
        product_root = self.x_scope.model('Product')
        self.x_target_model = product_root.add_model(name='Reference target', multiplicity=Multiplicity.ZERO_MANY)
        self.x_target = product_root.instance().add(model=self.x_target_model, name='Target instance')

    def tearDown(self):
        self.x_reference_model.delete()
        self.x_scope.delete()
        super(TestMultiReferencePropertyXScope, self).tearDown()

    def test_set_model_value(self):
        # setUp
        self.x_reference_model.value = [self.x_target_model]
        self.x_reference_model.refresh()

        # testing
        self.assertEqual(self.x_scope.id, self.x_reference_model.value[0].scope_id)
        self.assertEqual(self.x_target_model.id, self.x_reference_model.value[0].id)

    def test_set_model_value_using_id(self):
        # setUp
        self.x_reference_model.value = [self.x_target_model.id]
        self.x_reference_model.refresh()

        # testing
        self.assertEqual(self.x_scope.id, self.x_reference_model.value[0].scope_id)
        self.assertEqual(self.x_target_model.id, self.x_reference_model.value[0].id)

    def test_set_value(self):
        # setUp
        self.x_reference_model.value = [self.x_target_model.id]
        self.x_reference.refresh()
        self.x_reference.value = [self.x_target]

        self.assertTrue(len(self.x_reference.value) == 1)
        self.assertEqual(self.x_target.id, self.x_reference.value[0].id)



import os
from unittest import TestCase

import six
from betamax import Betamax

if six.PY2:
    from test.test_support import EnvironmentVarGuard
elif six.PY3:
    from test.support import EnvironmentVarGuard

from pykechain import Client
from tests.utils import TEST_TOKEN, TEST_SCOPE_NAME, TEST_URL, TEST_RECORD_CASSETTES, TEST_SCOPE_ID

with Betamax.configure() as config:
    config.cassette_library_dir = os.path.join(os.path.dirname(__file__), 'cassettes')
    config.define_cassette_placeholder('<API_URL>', TEST_URL)
    config.define_cassette_placeholder('<AUTH_TOKEN>', TEST_TOKEN)


class SixTestCase(TestCase):
    def assertRaisesRegex(self, expected_exception, expected_regex, *args, **kwargs):
        if six.PY2:
            return self.assertRaisesRegexp(expected_exception, expected_regex, *args, **kwargs)
        else:
            return super(__class__, self).assertRaisesRegex(expected_exception, expected_regex, *args, **kwargs)

    def assertRegex(self, text, expected_regex, *args, **kwargs):
        if six.PY2:
            return self.assertRegexpMatches(text, expected_regex, *args, **kwargs)
        else:
            return super(__class__, self).assertRegex(text, expected_regex, *args, **kwargs)


class TestBetamax(SixTestCase):
    @property
    def cassette_name(self):
        cls = getattr(self, '__class__')
        test = self._testMethodName
        return '{0}.{1}'.format(cls.__name__, test)

    def setUp(self):
        # use self.env.set('var', 'value') and with self.env: ... to use custom envvars
        self.env = EnvironmentVarGuard()

        self.client = Client(url=TEST_URL)

        if TEST_TOKEN:
            self.client.login(token=TEST_TOKEN)

        self.recorder = Betamax(session=self.client.session)

        if TEST_RECORD_CASSETTES:
            self.recorder.use_cassette(self.cassette_name)
            self.recorder.start()
        else:
            self.recorder.config.record_mode = None

        if TEST_SCOPE_NAME:
            self.project = self.client.scope(name=TEST_SCOPE_NAME)
        elif TEST_SCOPE_ID:
            self.project = self.client.scope(id=TEST_SCOPE_ID)
        else:
            raise Exception('Cloud not retrieve the test scope, you need to provide a '
                            '`TEST_SCOPE_ID` or `TEST_SCOPE_NAME` in the `.env` file')

    def tearDown(self):
        self.recorder.stop()
        del self.env
        del self.client



from pykechain.enums import PropertyRepresentation, SelectListRepresentations, LinkTargets
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.representations.representation_base import BaseRepresentation


class DecimalPlaces(BaseRepresentation):
    """Representation for floating-point value properties."""

    rtype = PropertyRepresentation.DECIMAL_PLACES
    _config_value_key = 'amount'

    def validate_representation(self, value):
        # type: (int) -> None
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: int
        :return: None
        """
        if not isinstance(value, int):
            raise IllegalArgumentError('{} value "{}" is not correct: not an integer'.format(
                self.__class__.__name__, value))


class SignificantDigits(DecimalPlaces):
    """Representation for floating-point value properties."""

    rtype = PropertyRepresentation.SIGNIFICANT_DIGITS


class ThousandsSeparator(BaseRepresentation):
    """Representation for integer or floating-point value properties."""

    rtype = PropertyRepresentation.THOUSANDS_SEPARATOR

    def validate_representation(self, value):
        # type: (None) -> None
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: int or float
        :return: None
        """
        if not (isinstance(value, type(None))):
            raise IllegalArgumentError('{} value "{}" is not correct: not NoneType'.format(
                self.__class__.__name__, value))


class LinkTarget(BaseRepresentation):
    """Representation for HTML link reference properties."""

    rtype = PropertyRepresentation.LINK_TARGET
    _config_value_key = 'target'

    def validate_representation(self, value):
        # type: (LinkTarget) -> None
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: int
        :return: None
        """
        if value not in LinkTargets.values():
            raise IllegalArgumentError('{} value "{}" is not correct: Not an CardWidgetLinkTarget option.'.format(
                self.__class__.__name__, value))


class ButtonRepresentation(BaseRepresentation):
    """Representation for single-select list properties."""

    rtype = PropertyRepresentation.BUTTON
    _config_value_key = PropertyRepresentation.BUTTON

    def validate_representation(self, value):
        # type: (SelectListRepresentations) -> None
        """
        Validate whether the representation value can be set.

        :param value: representation value to set.
        :type value: int
        :return: None
        """
        if value not in SelectListRepresentations.values():
            raise IllegalArgumentError('{} value "{}" is not correct: Not an SelectListRepresentations option.'.format(
                self.__class__.__name__, value))



from pykechain.models.widgets.widget import Widget
from pykechain.models.widgets.widget_schemas import undefined_meta_schema


# UNDEFINED = 'UNDEFINED'
# PROPERTYGRID = 'PROPERTYGRID'
# SUPERGRID = 'SUPERGRID'
# HTML = 'HTML'
# FILTEREDGRID = 'FILTEREDGRID'
# SERVICE = 'SERVICE'
# NOTEBOOK = 'NOTEBOOK'
# ATTACHMENTVIEWER = 'ATTACHMENTVIEWER'
# TASKNAVIGATIONBAR = 'TASKNAVIGATIONBAR'
# JSON = 'JSON'
# METAPANEL = 'METAPANEL'
# MULTICOLUMN = 'MULTICOLUMN'
# SCOPE_WIDGET = 'SCOPE_WIDGET'
# THIRD_PARTY = 'THIRD_PARTY'
# PROGRESS = 'PROGRESS'
# SIGNATURE = 'SIGNATURE'
# CARD = 'CARD'


class MetapanelWidget(Widget):
    """Metapanel Widget."""


class PropertygridWidget(Widget):
    """Propertygrid Widget."""


class UndefinedWidget(Widget):
    """Undefined Widget."""

    schema = undefined_meta_schema


class FilteredgridWidget(Widget):
    """Filteredgrid Widget."""


class SupergridWidget(Widget):
    """Supergrid Widget."""


class AttachmentviewerWidget(Widget):
    """Attachmentviewer Widget."""


class TasknavigationbarWidget(Widget):
    """Tasknavigationbar Widget."""


class HtmlWidget(Widget):
    """HTML Widget."""


class ServiceWidget(Widget):
    """Service Widget."""


class NotebookWidget(Widget):
    """Notebook Widget."""


class JsonWidget(Widget):
    """JSON Widget."""


class MulticolumnWidget(Widget):
    """Multicolumn Widget."""


class ProgressWidget(Widget):
    """Progress bar Widget."""


class ScopeWidget(Widget):
    """Scope grid Widget."""


class SignatureWidget(Widget):
    """Signature Widget."""


class CardWidget(Widget):
    """Card Widget."""


class ThirdpartyWidget(Widget):
    """Thirdparty Widget."""


class TasksWidget(Widget):
    """Tasks Widget."""



from six import text_type, string_types

from pykechain.models.part import Part
from pykechain.models.property import Property
from pykechain.utils import is_uuid


class MultiReferenceProperty(Property):
    """A virtual object representing a KE-chain multi-references property.

    .. versionadded:: 1.14
    """

    def __init__(self, json, **kwargs):
        """Construct a MultiReferenceProperty from a json object."""
        super(MultiReferenceProperty, self).__init__(json, **kwargs)

        self._cached_values = None

    @property
    def value(self):
        """Value of a reference property.

        You can set the reference with a Part, Part id or None value.
        Ensure that the model of the provided part, matches the configured model

        :return: a :class:`Part` or None
        :raises APIError: When unable to find the associated :class:`Part`

        Example
        -------
        Get the wheel reference property

        >>> part = project.part('Bike')
        >>> wheels_ref_property = part.property('Wheels')
        >>> isinstance(wheels_ref_property, MultiReferenceProperty)
        True

        The value returns a list of Parts or is an empty list

        >>> type(wheels_ref_property.value) in (list, tuple)
        True

        Get the selection of wheel instances:

        >>> wheel_choices = wheels_ref_property.choices()

        Choose random wheel from the wheel_choices:

        >>> from random import choice
        >>> wheel_choice_1 = choice(wheel_choices)
        >>> wheel_choice_2 = choice(wheel_choices)

        Set chosen wheel
        1: provide a single wheel:

        >>> wheels_ref_property.value = [wheel_choice_1]

        2: provide multiple wheels:

        >>> wheels_ref_property.value = [wheel_choice_1, wheel_choice_2]

        """
        if not self._value:
            return None
        if not self._cached_values and isinstance(self._value, (list, tuple)):
            ids = [v.get('id') for v in self._value]
            self._cached_values = list(self._client.parts(id__in=','.join(ids), category=None))
        return self._cached_values

    @value.setter
    def value(self, value):
        value_to_set = []
        if isinstance(value, (list, tuple)):
            for item in value:
                if isinstance(item, Part):
                    value_to_set.append(item.id)
                elif isinstance(item, (string_types, text_type)) and is_uuid(item):
                    # tested against a six.text_type (found in the requests' urllib3 package) for unicode
                    # conversion in py27
                    value_to_set.append(item)
                else:
                    raise ValueError("Reference must be a Part, Part id or None. type: {}".format(type(item)))
        elif isinstance(value, type(None)):
            # clear out the list
            value_to_set = None
        else:
            raise ValueError(
                "Reference must be a list (or tuple) of Part, Part id or None. type: {}".format(type(value)))

        # we replace the current choices!
        self._value = self._put_value(value_to_set)

    def choices(self):
        """Retrieve the parts that you can reference for this `MultiReferenceProperty`.

        This method makes 2 API calls: 1) to retrieve the referenced model, and 2) to retrieve the instances of
        that model.

        :return: the :class:`Part`'s that can be referenced as a :class:`~pykechain.model.PartSet`.
        :raises APIError: When unable to load and provide the choices

        Example
        -------
        >>> property = project.part('Bike').property('a_multi_reference_property')
        >>> reference_part_choices = property.choices()

        """
        # from the reference property (instance) we need to get the value of the reference property in the model
        # in the reference property of the model the value is set to the ID of the model from which we can choose parts
        model_parent_part = self.part.model()  # makes single part call
        property_model = model_parent_part.property(self.name)
        referenced_model = property_model.value and property_model.value[0]  # list of seleceted models is always 1
        possible_choices = self._client.parts(model=referenced_model)  # makes multiple parts call

        return possible_choices



import datetime

from pykechain.enums import PropertyType
from pykechain.exceptions import IllegalArgumentError
from pykechain.models.property2_datetime import DatetimeProperty2
from tests.classes import TestBetamax


class TestDateTime(TestBetamax):
    example_time = datetime.datetime.now()

    def setUp(self) -> None:
        super(TestDateTime, self).setUp()

        self.property_model = self.project.model('Bike').add_property(
            name='Testing date', property_type=PropertyType.DATETIME_VALUE)  # type: DatetimeProperty2
        self.property = self.project.part('Bike').property(name='Testing date')  # type: DatetimeProperty2

        self.property.value = self.example_time

    def tearDown(self):
        self.property_model.delete()
        super(TestDateTime, self).tearDown()

    def test_get_datetime(self):
        value = self.property.value

        self.assertIsInstance(value, str)

    def test_set_value_none(self):
        self.property.value = None

    def test_set_value_non_datetime(self):
        with self.assertRaises(IllegalArgumentError):
            self.property.value = '3'

    def test_set_value(self):
        self.property.value = self.example_time

    def test_to_datetime(self):
        date_time_value = self.property.to_datetime()

        self.assertIsInstance(date_time_value, datetime.datetime)



import datetime
import warnings
from typing import Any  # noqa: F401

import requests
from six import text_type, string_types

from pykechain.enums import Multiplicity, ScopeStatus
from pykechain.exceptions import APIError, NotFoundError, IllegalArgumentError, ForbiddenError
from pykechain.models.base import Base
from pykechain.models.team import Team
from pykechain.utils import is_uuid, parse_datetime


class Scope(Base):
    """A virtual object representing a KE-chain scope.

    :ivar name: Name of the scope
    :type name: str
    :ivar id: UUID of the scope
    :type if: uuid
    :ivar workflow_root: Root of the workflow
    :type workflow_root: uuid
    :ivar description: Description of the scope
    :type description: str or None
    :ivar status: Status of the scope, enumeration of `ScopeStatus`
    :type status: str
    :ivar start_date: Start date of the scope
    :type start_date: datetime or None
    :ivar due_date: Due date of the scope
    :type due_date: datetime or None
    :ivar type: Type of scope, enumeration of `ScopeType`
    :type type: str
    :ivar tags: list of unique tags of the scope.
    :type tags: list of str
    :ivar team: Team associated to the scope
    :type team: `Team` or None
    """

    def __init__(self, json, **kwargs):
        # type: (dict, **Any) -> None
        """Construct a scope from provided json data."""
        super(Scope, self).__init__(json, **kwargs)

        self.bucket = json.get('bucket', {})

        # for 'kechain2.core.wim <2.0.0'
        self.process = json.get('process')
        # for 'kechain2.core.wim >=2.0.0'
        self.workflow_root = json.get('workflow_root_id')

        self.description = json.get('description')
        self.status = json.get('status')
        self.type = json.get('type')

        self.start_date = parse_datetime(json.get('start_date'))
        self.due_date = parse_datetime(json.get('due_date'))

    def __repr__(self):  # pragma: no cover
        return "<pyke Scope '{}' id {}>".format(self.name, self.id[-8:])

    @property
    def team(self):
        """Team to which the scope is assigned."""
        team_dict = self._json_data.get('team')
        if team_dict and team_dict.get('id'):
            return self._client.team(pk=team_dict.get('id'))
        else:
            return None

    @property
    def tags(self):
        """
        Tags of the scope.

        A list of unique strings.

        :return: list of tag strings
        :rtype: list of str
        """
        return self._json_data.get('tags')

    def parts(self, *args, **kwargs):
        """Retrieve parts belonging to this scope.

        See :class:`pykechain.Client.parts` for available parameters.
        """
        return self._client.parts(*args, bucket=self.bucket.get('id'), **kwargs)

    def part(self, *args, **kwargs):
        """Retrieve a single part belonging to this scope.

        See :class:`pykechain.Client.part` for available parameters.
        """
        return self._client.part(*args, bucket=self.bucket.get('id'), **kwargs)

    def create_model(self, parent, name, multiplicity=Multiplicity.ZERO_MANY):
        """Create a single part model in this scope.

        See :class:`pykechain.Client.create_model` for available parameters.
        """
        return self._client.create_model(parent, name, multiplicity=multiplicity)

    def model(self, *args, **kwargs):
        """Retrieve a single model belonging to this scope.

        See :class:`pykechain.Client.model` for available parameters.
        """
        return self._client.model(*args, bucket=self.bucket.get('id'), **kwargs)

    def activities(self, *args, **kwargs):
        """Retrieve activities belonging to this scope.

        See :class:`pykechain.Client.activities` for available parameters.
        """
        if self._client.match_app_version(label='wim', version='<2.0.0', default=True):
            return self._client.activities(*args, scope=self.id, **kwargs)
        else:
            return self._client.activities(*args, scope_id=self.id, **kwargs)

    def activity(self, *args, **kwargs):
        """Retrieve a single activity belonging to this scope.

        See :class:`pykechain.Client.activity` for available parameters.
        """
        if self._client.match_app_version(label='wim', version='<2.0.0', default=True):
            return self._client.activity(*args, scope=self.id, **kwargs)
        else:
            return self._client.activity(*args, scope_id=self.id, **kwargs)

    def create_activity(self, *args, **kwargs):
        """Create a new activity belonging to this scope.

        See :class:`pykechain.Client.create_activity` for available parameters.
        """
        if self._client.match_app_version(label='wim', version='<2.0.0', default=True):
            return self._client.create_activity(self.process, *args, **kwargs)
        else:
            return self._client.create_activity(self.workflow_root, *args, **kwargs)

    def services(self, *args, **kwargs):
        """Retrieve services belonging to this scope.

        See :class:`pykechain.Client.services` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.services(*args, scope=self.id, **kwargs)

    def create_service(self, *args, **kwargs):
        """Create a service to current scope.

        See :class:`pykechain.Client.create_service` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.create_service(*args, scope=self.id, **kwargs)

    def service(self, *args, **kwargs):
        """Retrieve a single service belonging to this scope.

        See :class:`pykechain.Client.service` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service(*args, scope=self.id, **kwargs)

    def service_executions(self, *args, **kwargs):
        """Retrieve services belonging to this scope.

        See :class:`pykechain.Client.service_executions` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service_executions(*args, scope=self.id, **kwargs)

    def service_execution(self, *args, **kwargs):
        """Retrieve a single service execution belonging to this scope.

        See :class:`pykechain.Client.service_execution` for available parameters.

        .. versionadded:: 1.13
        """
        return self._client.service_execution(*args, scope=self.id, **kwargs)

    def members(self, is_manager=None):
        """
        Retrieve members of the scope.

        :param is_manager: (optional) set to True to return only Scope members that are also managers.
        :type is_manager: bool
        :return: List of members (usernames)

        Examples
        --------
        >>> members = project.members()
        >>> managers = project.members(is_manager=True)

        """
        if not is_manager:
            return [member for member in self._json_data['members'] if member['is_active']]
        else:
            return [member for member in self._json_data['members'] if
                    member.get('is_active', False) and member.get('is_manager', False)]

    def add_member(self, member):
        """
        Add a single member to the scope.

        You may only edit the list of members if the pykechain credentials allow this.

        :param member: single username to be added to the scope list of members
        :type member: basestring
        :raises APIError: when unable to update the scope member
        """
        select_action = 'add_member'

        self._update_scope_project_team(select_action=select_action, user=member, user_type='member')

    def remove_member(self, member):
        """
        Remove a single member to the scope.

        :param member: single username to be removed from the scope list of members
        :type member: basestring
        :raises APIError: when unable to update the scope member
        """
        select_action = 'remove_member'

        self._update_scope_project_team(select_action=select_action, user=member, user_type='member')

    def add_manager(self, manager):
        """
        Add a single manager to the scope.

        :param manager: single username to be added to the scope list of managers
        :type manager: basestring
        :raises APIError: when unable to update the scope manager
        """
        select_action = 'add_manager'

        self._update_scope_project_team(select_action=select_action, user=manager, user_type='manager')

    def remove_manager(self, manager):
        """
        Remove a single manager to the scope.

        :param manager: single username to be added to the scope list of managers
        :type manager: basestring
        :raises APIError: when unable to update the scope manager
        """
        select_action = 'remove_manager'

        self._update_scope_project_team(select_action=select_action, user=manager, user_type='manager')

    def _update_scope_project_team(self, select_action, user, user_type):
        """
        Update the Project Team of the Scope. Updates include addition or removing of managers or members.

        :param select_action: type of action to be applied
        :type select_action: basestring
        :param user: the username of the user to which the action applies to
        :type user: basestring
        :param user_type: the type of the user (member or manager)
        :type user_type: basestring
        :raises APIError: When unable to update the scope project team.
        """
        if isinstance(user, (string_types, text_type)):
            users = self._client._retrieve_users()
            manager_object = next((item for item in users['results'] if item["username"] == user), None)
            if manager_object:
                url = self._client._build_url('scope', scope_id=self.id)
                response = self._client._request('PUT', url, params={'select_action': select_action},
                                                 data={'user_id': manager_object['pk']})
                if response.status_code != requests.codes.ok:  # pragma: no cover
                    raise APIError("Could not {} {} in Scope".format(select_action.split('_')[0], user_type))
            else:
                raise NotFoundError("User {} does not exist".format(user))
        else:
            raise TypeError("User {} should be defined as a string".format(user))

    def _edit(self, update_dict):
        url = self._client._build_url('scope', scope_id=self.id)

        response = self._client._request('PUT', url, json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Scope ({})".format(response))
        else:
            self._json_data = response.json().get('results') and response.json().get('results')[0]

    def edit(self, name=None, description=None, start_date=None, due_date=None, status=None, tags=None, team=None,
             options=None, **kwargs):
        """Edit the details of a scope.

        :param name: (optionally) edit the name of the scope
        :type name: basestring or None
        :param description: (optionally) edit the description of the scope
        :type description: basestring or None
        :param start_date: (optionally) edit the start date of the scope as a datetime object (UTC time/timezone
                            aware preferred)
        :type start_date: datetime or None
        :param due_date: (optionally) edit the due_date of the scope as a datetime object (UTC time/timzeone
                            aware preferred)
        :type due_date: datetime or None
        :param status: (optionally) edit the status of the scope as a string based
        :type status: basestring or None
        :param tags: (optionally) replace the tags on a scope, which is a list of strings ["one","two","three"]
        :type tags: list of basestring or None
        :param team: (optionally) add the scope to a team
        :type team: UUIDstring or None
        :param options: (optionally) custom options dictionary stored on the scope object
        :type options: dict or None
        :raises IllegalArgumentError: if the type of the inputs is not correct
        :raises APIError: if another Error occurs
        :warns: UserWarning - When a naive datetime is provided. Defaults to UTC.

        Examples
        --------
        >>> from datetime import datetime
        >>> project.edit(name='New project name',
        ...              description='Changing the description just because I can',
        ...              start_date=datetime.utcnow(),  # naive time is interpreted as UTC time
        ...              status=ScopeStatus.CLOSED)

        If we want to provide timezone aware datetime objects we can use the 3rd party convenience library :mod:`pytz`.
        Mind that we need to fetch the timezone first and use `<timezone>.localize(<your datetime>)` to make it
        work correctly.

        Using `datetime(2017,6,1,23,59,0 tzinfo=<tz>)` does NOT work for most timezones with a
        daylight saving time. Check the `pytz <http://pythonhosted.org/pytz/#localized-times-and-date-arithmetic>`_
        documentation.

        To make it work using :mod:`pytz` and timezone aware :mod:`datetime` see the following example::

        >>> import pytz
        >>> start_date_tzaware = datetime.now(pytz.utc)
        >>> mytimezone = pytz.timezone('Europe/Amsterdam')
        >>> due_date_tzaware = mytimezone.localize(datetime(2019, 10, 27, 23, 59, 0))
        >>> project.edit(due_date=due_date_tzaware, start_date=start_date_tzaware)

        To assign a scope to a team see the following example::

        >>> my_team = client.team(name='My own team')
        >>> project.edit(team=my_team)

        """
        update_dict = {'id': self.id}
        if name is not None:
            if isinstance(name, (str, text_type)):
                update_dict.update({'name': name})
                self.name = name
            else:
                raise IllegalArgumentError('Name should be a string')

        if description is not None:  # isinstance(description, (str, text_type)):
            if isinstance(description, (str, text_type)):
                update_dict.update({'text': description})
                self.text = description
            else:
                raise IllegalArgumentError('Description should be a string')

        if start_date is not None:
            if isinstance(start_date, datetime.datetime):
                if not start_date.tzinfo:
                    warnings.warn("The startdate '{}' is naive and not timezone aware, use pytz.timezone info. "
                                  "This date is interpreted as UTC time.".format(start_date.isoformat(sep=' ')))
                update_dict.update({'start_date': start_date.isoformat(sep='T')})
            else:
                raise IllegalArgumentError('Start date should be a datetime.datetime() object')

        if due_date is not None:
            if isinstance(due_date, datetime.datetime):
                if not due_date.tzinfo:
                    warnings.warn("The duedate '{}' is naive and not timezone aware, use pytz.timezone info. "
                                  "This date is interpreted as UTC time.".format(due_date.isoformat(sep=' ')))
                update_dict.update({'due_date': due_date.isoformat(sep='T')})
            else:
                raise IllegalArgumentError('Due date should be a datetime.datetime() object')

        if status is not None:
            if isinstance(status, (str, text_type)) and status in ScopeStatus.values():
                update_dict.update({'status': status})
            else:
                raise IllegalArgumentError('Status should be a string and in the list of acceptable '
                                           'status strings: {}'.format(ScopeStatus.values()))

        if tags is not None:
            if isinstance(tags, (list, tuple, set)):
                update_dict.update({'tags': tags})
            else:
                raise IllegalArgumentError('tags should be a an array (list, tuple, set) of strings')

        if team is not None:
            if isinstance(team, (str, text_type)) and is_uuid(team):
                update_dict.update({'team_id': team})
            elif isinstance(team, Team):
                update_dict.update({'team_id': team.id})
            else:
                raise IllegalArgumentError("team should be the uuid of a team")

        if options is not None:
            if isinstance(options, dict):
                update_dict.update({'options': options})
            else:
                raise IllegalArgumentError("options should be a dictionary")

        # do the update itself in an abstracted function.
        self._edit(update_dict)

    def clone(self, *args, **kwargs):
        """
        Clone current scope.

        See :class:`pykechain.Client.clone_scope` for available parameters.

        .. versionadded:: 2.6.0
        """
        return self._client.clone_scope(*args, source_scope=self, **kwargs)

    def delete(self):
        """
        Delete the scope.

        Only works with enough permissions.

        .. versionadded: 3.0

        :raises ForbiddenError: if you do not have the permissions to delete a scope
        """
        url = self._client._build_url('scope', scope_id=self.id)
        response = self._client._request('DELETE', url)

        if response.status_code != requests.codes.no_content:  # pragma: no cover
            if response.status_code == requests.codes.forbiddem:
                raise ForbiddenError("Forbidden to delete scope, {}: {}".format(str(response), response.content))
            raise APIError("Could not delete scope, {}: {}".format(str(response), response.content))



from six import text_type

from pykechain.models.part import Part
from pykechain.models.property import Property


class ReferenceProperty(Property):
    """A virtual object representing a KE-chain reference property."""

    @property
    def value(self):
        """Value of a reference property.

        You can set the reference with a Part, Part id or None value.
        Ensure that the model of the provided part, matches the configured model

        :return: a :class:`Part` or None
        :raises APIError: When unable to find the associated :class:`Part`

        Example
        -------
        Get the wheel reference property

        >>> part = project.part('Bike')
        >>> material_ref_property = part.property('Material Selection')
        >>> isinstance(material_ref_property, ReferenceProperty)
        True

        The value either returns a Part or is None if not set (yet)

        >>> type(material_ref_property.value) in (Part, None)
        True

        Get the selection of material instances containing the word material (icontains is for case-insensitive)

        >>> material_choices = project.parts(icontains='material')

        Choose random material option

        >>> from random import choice
        >>> chosen_material = choice(material_choices)

        Set chosen material
        1: provide the part

        >>> material_ref_property.value = chosen_material

        2: provide the id

        >>> material_ref_property.value = chosen_material.id

        """
        if not self._value:
            return None

        return self._client.part(pk=self._value['id'])

    @value.setter
    def value(self, value):
        if isinstance(value, Part):
            part_id = value.id
        elif isinstance(value, (type(None), text_type)):
            # tested against a six.text_type (found in the requests' urllib3 package) for unicode conversion in py27
            part_id = value
        else:
            raise ValueError("Reference must be a Part, Part id or None. type: {}".format(type(value)))

        self._value = self._put_value(part_id)

    def choices(self):
        """Retrieve the parts that you can reference for this `ReferenceProperty`.

        This method makes 2 API calls: 1) to retrieve the referenced model, and 2) to retrieve the instances of
        that model.

        :return: the :class:`Part`'s that can be referenced as a :class:`~pykechain.model.PartSet`.
        :raises APIError: When unable to load and provide the choices

        Example
        -------
        >>> property = project.part('Bike').property('RefTest')
        >>> reference_part_choices = property.choices()

        """
        # from the reference property (instance) we need to get the value of the reference property in the model
        # in the reference property of the model the value is set to the ID of the model from which we can choose parts
        model_parent_part = self.part.model()  # makes single part call
        property_model = model_parent_part.property(self.name)
        referenced_model = self._client.model(pk=property_model._value['id'])  # makes single part call
        possible_choices = self._client.parts(model=referenced_model)  # makes multiple parts call

        return possible_choices



from jsonschema import validate

from pykechain.enums import PropertyType, Multiplicity, LinkTargets, SelectListRepresentations
from pykechain.exceptions import IllegalArgumentError
from pykechain.models import AnyProperty
from pykechain.models.representations.representation_base import BaseRepresentation
from pykechain.models.representations.representations import ButtonRepresentation, LinkTarget, SignificantDigits, \
    DecimalPlaces, ThousandsSeparator
from pykechain.models.validators.validator_schemas import options_json_schema
from tests.classes import SixTestCase, TestBetamax


class TestRepresentationJSON(SixTestCase):
    def test_valid_button_representation_json(self):
        options = dict(
            representations=[dict(
                rtype='buttonRepresentation',
                config=dict(
                    buttonRepresentation="dropdown",
                )
            )]
        )
        validate(options_json_schema, options)


class TestSignificantDigitsLive(TestBetamax):
    """
    Specify these class attributes in every sub-class
    """
    property_type = PropertyType.FLOAT_VALUE  # type: PropertyType
    representation_class = SignificantDigits  # type: type(BaseRepresentation)
    value = 3
    new_value = 1

    test_model_name = '__test model for representations'
    test_prop_name = '__test property for representations'

    def setUp(self):
        super().setUp()
        product_root = self.project.model(name='Product')
        self.test_model = product_root.add_model(name=self.test_model_name, multiplicity=Multiplicity.ONE)
        self.prop_model = self.test_model.add_property(
            name=self.test_prop_name,
            property_type=self.property_type,
        )  # type: AnyProperty

    def tearDown(self):
        self.test_model.delete()
        super().tearDown()

    def test_getter(self):
        representations = self.prop_model.representations

        self.assertIsInstance(representations, list)
        self.assertTrue(all([isinstance(r, self.representation_class) for r in representations]))

    def test_setter(self):
        self.prop_model.representations = [
            self.representation_class(
                prop=self.prop_model,
                value=self.value,
            ),
        ]

        model = self.project.model(name=self.test_model_name)
        first_representation = model.property(name=self.test_prop_name).representations[0]

        self.assertIsInstance(first_representation, self.representation_class)
        self.assertTrue(first_representation.value == self.value)

    def test_set_value(self):
        # Make sure there is a representation object
        self.prop_model.representations = [
            self.representation_class(
                prop=self.prop_model,
                value=self.value,
            ),
        ]

        # Set new value
        first_repr = self.prop_model.representations[0]
        first_repr.value = self.new_value

        # Retrieve representation again
        model = self.project.model(name=self.test_model_name)
        first_representation = model.property(name=self.test_prop_name).representations[0]

        # testing
        self.assertEqual(first_repr.value, first_representation.value)

    def test_unsupported_value(self):

        with self.assertRaises(IllegalArgumentError):
            self.representation_class(
                prop=self.prop_model,
                value='Unexpected value',
            )


class TestThousandsSeparatorLive(TestSignificantDigitsLive):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = ThousandsSeparator
    value = None
    new_value = None


class TestDecimalPlacesLive(TestSignificantDigitsLive):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = DecimalPlaces
    value = 3
    new_value = 5


class TestLinkTargetLive(TestSignificantDigitsLive):
    property_type = PropertyType.LINK_VALUE
    representation_class = LinkTarget
    value = LinkTargets.SAME_TAB
    new_value = LinkTargets.NEW_TAB


class TestButtonRepresentationLive(TestSignificantDigitsLive):
    property_type = PropertyType.SINGLE_SELECT_VALUE
    representation_class = ButtonRepresentation
    value = SelectListRepresentations.CHECK_BOXES
    new_value = SelectListRepresentations.BUTTONS

    def setUp(self):
        super().setUp()
        self.prop_model.options = ['alpha', 'beta', 'gamma', 'omega']
